"""

    Greynir: Natural language processing for Icelandic

    Python wrapper for C++ Earley/Scott parser

    Copyright (C) 2021 Miðeind ehf.

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    This module wraps an Earley-Scott parser written in C++ to transform token
    sequences (sentences) into forests of parse trees, with each tree representing a
    possible parse of a sentence, according to a given context-free grammar.

    An Earley parser handles all valid context-free grammars,
    irrespective of ambiguity, recursion (left/middle/right), nullability, etc.
    The returned parse trees reflect the original grammar, which does not
    need to be normalized or modified in any way. All partial parses are
    available in the final parse state table.

    For further information see J. Earley, "An efficient context-free parsing algorithm",
    Communications of the Association for Computing Machinery, 13:2:94-102, 1970.

    The Earley parser used here is the improved version described by Scott & Johnstone,
    referencing Tomita. This allows worst-case cubic (O(n^3)) order, where n is the
    length of the input sentence, while still returning all possible parse trees
    for an ambiguous grammar.

    See Elizabeth Scott, Adrian Johnstone:
    "Recognition is not parsing — SPPF-style parsing from cubic recognisers"
    Science of Computer Programming, Volume 75, Issues 1–2, 1 January 2010, Pages 55–70

    The C++ source code is found in eparser.h and eparser.cpp.

    This wrapper uses the CFFI module (http://cffi.readthedocs.org/en/latest/)
    to call C++ code from CPython and PyPy, also allowing the C++ code to
    call back into Python.

"""

from typing import (
    Dict,
    Any,
    Optional,
    List,
    Set,
    Union,
    Tuple,
    Iterable,
    Iterator,
    Hashable,
    IO,
    cast,
)

import os
import operator
from threading import Lock
from functools import reduce

from .binparser import (
    BIN_Parser,
    BIN_Token,
    simplify_terminal,
    augment_terminal,
    Tok,
    TokenDict,
)
from .grammar import Grammar, GrammarError, Nonterminal, Terminal, Production
from .settings import Settings
from .glock import GlobalLock

# Import the CFFI wrapper module for the _eparser.*.so library
# which is compiled from eparser.cpp (see eparser_build.py)
# pylint: disable=no-name-in-module
from ._eparser import lib as eparser, ffi  # type: ignore


ffi_NULL: Any = cast(Any, ffi).NULL

# The type of an entry on a ParseTreeFlattener stack
FlattenerType = Union[Tuple[Terminal, BIN_Token], Nonterminal]
ProductionTuple = Tuple[Production, List[Optional["Node"]]]


class ParseJob:

    """ Dispatch token matching requests coming in from the C++ code """

    # Parse jobs have rotating integer IDs, reaching _MAX_JOBS before cycling back
    _MAX_JOBS = 10_000
    _seq = 0
    _jobs: Dict[int, "ParseJob"] = dict()
    _lock = Lock()

    def __init__(
        self,
        handle: int,
        grammar: Grammar,
        tokens: List[BIN_Token],
        terminals: Dict[int, Terminal],
        matching_cache: Dict[Tuple[Hashable, ...], Any],
    ) -> None:
        self._handle = handle
        self.tokens = tokens
        self.terminals = terminals
        self.grammar = grammar
        self.c_dict: Dict[Any, "Node"] = dict()  # Node pointer conversion dictionary
        self.matching_cache = matching_cache  # Token/terminal matching buffers

    def matches(self, token_index: int, terminal_index: int) -> bool:
        """ Convert the token reference from a 0-based token index
            to the token object itself; convert the terminal from a
            1-based terminal index to a terminal object. """
        return self.tokens[token_index].matches(self.terminals[terminal_index])

    def alloc_cache(self, token: int, size: int) -> Any:
        """ Allocate a token/terminal matching cache buffer for the given token """
        key = self.tokens[token].key  # Obtain the (hashable) key of the BIN_Token
        try:
            # Do we already have a token/terminal cache match buffer for this key?
            b: Any = self.matching_cache.get(key)
            if b is None:
                # No: create a fresh one (assumed to be initialized to zero)
                b = self.matching_cache[key] = ffi.new("BYTE[]", size)  # type: ignore
        except TypeError:
            assert False, "alloc_cache() unable to hash key: {0}".format(repr(key))
        return b

    def reset(self) -> None:
        """ Reset the node pointer conversion dictionary """
        self.c_dict = dict()

    @property
    def handle(self) -> int:
        return self._handle

    def __enter__(self):
        """ Python context manager protocol """
        return self

    # noinspection PyUnusedLocal
    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any):
        """ Python context manager protocol """
        self.__class__.delete(self._handle)
        # Return False to re-throw exception from the context, if any
        return False

    @classmethod
    def make(
        cls,
        grammar: Grammar,
        tokens: List[BIN_Token],
        terminals: Dict[int, Terminal],
        matching_cache: Dict[Tuple[Hashable, ...], Any],
    ) -> "ParseJob":
        """ Create a new parse job with for a given token sequence and set of terminals """
        with cls._lock:
            h = cls._seq
            cls._seq += 1
            if cls._seq >= cls._MAX_JOBS:
                cls._seq = 0
            j = cls._jobs[h] = ParseJob(h, grammar, tokens, terminals, matching_cache)
        return j

    @classmethod
    def delete(cls, handle: int) -> None:
        """ Delete a no-longer-used parse job """
        with cls._lock:
            del cls._jobs[handle]

    @classmethod
    def dispatch(cls, handle: int, token_index: int, terminal_index: int) -> bool:
        """ Dispatch a match request to the correct parse job """
        return cls._jobs[handle].matches(token_index, terminal_index)

    @classmethod
    def alloc(cls, handle: int, token_index: int, size: int):
        """ Dispatch a cache buffer allocation request to the correct parse job """
        return cls._jobs[handle].alloc_cache(token_index, size)


# Declare CFFI callback functions to be called from the C++ code
# See: https://cffi.readthedocs.io/en/latest/using.html#extern-python-new-style-callbacks


@ffi.def_extern()  # type: ignore
def matching_func(handle: int, token_index: int, terminal_index: int) -> bool:
    """ This function is called from the C++ parser to determine
        whether a token matches a terminal. The token is referenced
        by 0-based index, and the terminal by a 1-based index.
        The handle is an arbitrary UINT that was passed to
        earleyParse(). In this case, it is used to identify
        a ParseJob object that dispatches the match query. """
    return ParseJob.dispatch(handle, token_index, terminal_index)


@ffi.def_extern()  # type: ignore
def alloc_func(handle: int, token_index: int, size: int):
    """ Allocate a token/terminal matching cache buffer, at least size bytes.
        If the callback returns ffi.NULL, the parser will allocate its own buffer.
        The point of this callback is to allow re-using buffers for identical tokens,
        so we avoid making unnecessary matching calls. """
    return ParseJob.alloc(handle, token_index, size)


class Node:

    """ Shared Packed Parse Forest (SPPF) node representation,
        mapped from C++ (see eparser.h/.cpp) to Python.

        A C++ node is described by a label tuple of
        (iNt, nI, nJ, nDot, pProd).

        nI and nJ are the start and end token indices of the span
        covered by the node.

        If iNt is >= 0, the node describes a token/terminal match
        and has no children. iNt is the index of the matched token.

        If iNt is < 0, it is the index of a nonterminal within the
        grammar being parsed. If pProd is not ffi.NULL, this is an
        interior node, corresponding to a dot position of nDot within
        the production pProd. Otherwise, i.e. if pProd is ffi.NULL,
        the node describes a completed nonterminal.

        Nonterminals, unless they are epsilon (empty) nodes, have
        a family of child (derivative) productions represented as a
        linked list starting with pHead. Since the SPPF trees coming
        from the Earley-Scott parser are binarized, the families have
        at most two child nodes each. When creating the Python nodes,
        interior nodes are coalesced as far as possible to form longer
        lists of children, thereby decreasing the total number of nodes
        necessary to represent the forest.

        A forest of Nodes can be navigated using a subclass of
        ParseForestNavigator.

    """

    # Note: __slots__ are not really required on PyPy,
    # but they are beneficial on CPython
    __slots__ = (
        "_start",
        "_end",
        "_families",
        "_highest_prio",
        "_nonterminal",
        "_terminal",
        "_token",
        "_completed",
        "score",
    )

    def __init__(self, start: int, end: int) -> None:
        # Start and end token indices, i.e. the span of this node
        self._start = start
        self._end = end
        # Families of children of this node
        self._families: Optional[List[ProductionTuple]] = None
        # Priority of highest-priority child family
        self._highest_prio = 0
        # The nonterminal corresponding to this node, if not a leaf node
        self._nonterminal: Optional[Nonterminal] = None
        # The terminal corresponding to this node, if it is a leaf node
        self._terminal: Optional[Terminal] = None
        # The token matching this terminal, if this is a leaf node
        self._token: Optional[BIN_Token] = None
        # If completed is True, this node represents a completed nonterminal.
        # Otherwise, it is an internal node representing a position within
        # a production of a nonterminal.
        self._completed = True
        # The score for this node, calculated by the reducer
        self.score = 0

    @classmethod
    def from_c_node(
        cls, job: ParseJob, c_node: Any, parent: Any = None, index: int = 0
    ) -> Optional["Node"]:
        """ Initialize a Python node from a C++ SPPF node structure """
        if c_node == ffi_NULL:
            return None

        lb = c_node.label
        if lb.nI >= lb.nJ:
            # Empty node (no tokens matched within it):
            # don't bother creating a corresponding Python object;
            # we never use these nodes anyway except as fillers
            # to help in matching children with the parent production
            return None

        node = cls(lb.nI, lb.nJ)  # Start token index, end token index

        if lb.iNt >= 0:
            # Token node: find the corresponding terminal
            assert parent is not None
            tix: int = parent.pList[index]
            node._terminal = job.grammar.lookup_terminal(tix)
            node._token = job.tokens[lb.iNt]
            return node

        # Nonterminal node
        nt = lb.iNt
        node._nonterminal = job.grammar.lookup_nonterminal(nt)
        node._completed = lb.pProd == ffi_NULL
        # Cache nonterminal nodes
        job.c_dict[c_node] = node

        # Loop through the families of children of this node
        fe = c_node.pHead
        while fe != ffi_NULL:

            # Save on node count by coalescing interior nodes
            # into the child list of the enclosing completed
            # nonterminal. Nodes can be coalesced while they
            # refer to the same nonterminal, are interior,
            # and not ambiguous.
            ch: List[Any] = []

            def push_pair(p1: Any, p2: Any) -> None:
                """ Push a pair of child nodes onto the child list """

                def push_child(p: Any) -> None:
                    """ Push a single child node onto the child list """
                    if p.label.iNt == nt and p.label.pProd != ffi_NULL:
                        # Interior node for the same nonterminal
                        if p.pHead.pNext == ffi_NULL:
                            # Unambiguous: recurse
                            push_pair(p.pHead.p1, p.pHead.p2)
                        else:
                            # Ambiguous node, i.e. more than one family of children.
                            # In this case we don't know which (p1,p2) pair
                            # to add as a child of the parent, so we must
                            # retain the original node with its family of children
                            # and end the recursion. We also need to add
                            # placeholder (dummy) nodes to keep the child
                            # list in sync with the nonterminal's production.
                            if p.label.nDot > 2:
                                # Add placeholders for the part of the production
                                # that is missing from the front since we abandon
                                # the recursion here
                                ch.extend([ffi_NULL] * (p.label.nDot - 2))
                            ch.append(p)
                            ch.append(ffi_NULL)  # Placeholder
                    else:
                        # Terminal, epsilon or unrelated nonterminal
                        ch.append(p)

                if p1 != ffi_NULL and p2 != ffi_NULL:
                    push_child(p1)
                    push_child(p2)
                elif p2 != ffi_NULL:
                    push_child(p2)
                else:
                    push_child(p1)

            push_pair(fe.p1, fe.p2)
            node._add_family(job, fe.pProd, ch)
            fe = fe.pNext

        return node

    @classmethod
    def copy(cls, other: "Node") -> "Node":
        """ Returns a copy of a Node instance """
        node = cls(other._start, other._end)
        node._nonterminal = other._nonterminal
        node._terminal = other._terminal
        node._token = other._token
        node._completed = other._completed
        node._highest_prio = other._highest_prio
        if other._families:
            # Create a new list object having the
            # same child nodes as the source node
            node._families = other._families[:]
        return node

    def _add_family(self, job: ParseJob, c_prod: Any, c_children: Any) -> None:
        """ Add a family of children to this node, in parallel with other families """
        assert c_prod != ffi_NULL
        prod: Production = job.grammar.productions_by_ix[c_prod.nId]
        prio: int = prod.priority
        # Note: lower priority values mean higher priority!
        if self._families and prio > self._highest_prio:
            # Lower priority than a family we already have: don't bother adding it
            return
        # Recreate the pc tuple from the production index
        pc: ProductionTuple = (
            prod,
            [
                # Convert child node from C++ form to Python form
                job.c_dict.get(ch) or Node.from_c_node(job, ch, c_prod, ix)
                for ix, ch in enumerate(c_children)
            ],
        )
        if self._families is None or prio < self._highest_prio:
            # First family of children, or highest priority so far: add as the only family
            self._families = [pc]
        else:
            # Same priority as the families we have already: add this
            self._families.append(pc)
        self._highest_prio = prio

    @property
    def start(self) -> int:
        """ Return the start token index """
        return self._start

    @property
    def end(self) -> int:
        """ Return the end token index """
        return self._end

    @property
    def is_span(self) -> bool:
        """ Returns True if the node spans one or more tokens """
        return self._end > self._start

    def _first_token(self) -> BIN_Token:
        """ Return the first token within the span of this node """
        p = self
        while p._token is None:
            # Note that this function may be called before the
            # tree is reduced; therefore we may have more than one
            # family of children. The families should however all
            # cover the same token span.
            assert p._families
            _, f = p._families[0]
            ix = 0
            while f[ix] is None:
                ix += 1
            assert f[ix] is not None
            p = cast(Node, f[ix])
        return p._token

    def _last_token(self) -> BIN_Token:
        """ Return the last token within the span of this node """
        p = self
        while p._token is None:
            # Note that this function may be called before the
            # tree is reduced; therefore we may have more than one
            # family of children. The families should however all
            # cover the same token span.
            assert p._families
            _, f = p._families[0]
            ix = -1
            while f[ix] is None:
                ix -= 1
            assert f[ix] is not None
            p = cast(Node, f[ix])
        return p._token

    @property
    def token_span(self) -> Tuple[BIN_Token, BIN_Token]:
        """ Return the first and last tokens under this node """
        return (self._first_token(), self._last_token())

    @property
    def nonterminal(self) -> Optional[Nonterminal]:
        """ Return the nonterminal associated with this node """
        return self._nonterminal

    @property
    def is_ambiguous(self):
        """ Return True if this node has more than one family of children """
        return self._families is not None and len(self._families) >= 2

    @property
    def is_interior(self) -> bool:
        """ Returns True if this is an interior node (partially parsed production) """
        return not self._completed

    @property
    def is_completed(self) -> bool:
        """ Returns True if this is a node corresponding to a completed nonterminal """
        return self._completed

    @property
    def is_token(self) -> bool:
        """ Returns True if this is a token node """
        return self._token is not None

    @property
    def terminal(self) -> Optional[Terminal]:
        """ Return the terminal associated with a token node, or None if none """
        return self._terminal

    @property
    def token(self) -> Optional[BIN_Token]:
        """ Return the token associated with a token node, or None if none """
        return self._token

    @property
    def has_children(self) -> bool:
        """ Return True if there are any families of children of this node """
        return bool(self._families)

    @property
    def is_empty(self) -> bool:
        """ Return True if there is only a single empty family of this node """
        if not self._families:
            return True
        return len(self._families) == 1 and not bool(self._families[0][1])

    @property
    def num_families(self) -> int:
        """ Return the number of families of children of this node """
        return len(self._families) if self._families is not None else 0

    def enum_children(self) -> Iterator[ProductionTuple]:
        """ Enumerate families of children """
        if self._families:
            for prod, children in self._families:
                yield (prod, children)

    def enum_child_nodes(self) -> Iterator[Optional["Node"]]:
        """ Enumerate child nodes of this node, one by one """
        # Note that for reduced trees, a nonterminal node
        # will only have one family of children, which
        # is the list that will be generated by this function.
        n = self.num_families
        if n == 0 or not self._families:
            # No children
            pass
        elif n == 1:
            # Single family of children: yield from that list
            _, children = self._families[0]
            yield from children
        else:
            # Multiple families of children,
            # meaning that this is an ambiguous, non-reduced node
            assert False, "enum_child_nodes() called on an ambiguous node"

    def reduce_to(self, child_ix: int) -> None:
        """ Eliminate all child families except the given one """
        if self._families and len(self._families) > 1:
            # More than one family to choose from:
            # collapse the list to one survivor
            f = self._families[child_ix]
            self._families = [f]

    def _repr(self, indent: int) -> str:
        if hasattr(self, "score"):
            sc = " [{0}] ".format(self.score)
        else:
            sc = ""
        if self._nonterminal is not None:
            label_rep = str(self._nonterminal) + sc
        else:
            label_rep = str(self._token) + " -> " + str(self._terminal) + sc
        families_rep = ""
        istr = "  " * indent
        if self._families:
            # Show the children in each family

            def child_rep(children: Iterable[Optional["Node"]]) -> str:
                if not children:
                    return ""
                return "\n".join(
                    ch._repr(indent + 1) for ch in children if ch is not None
                )

            if len(self._families) == 1:
                children = self._families[0][1]
                if not children:
                    families_rep = ""
                else:
                    families_rep = "\n" + child_rep(children)
            else:
                families_rep = "\n" + "\n".join(
                    istr + "Option " + str(ix + 1) + ":\n" + child_rep(child)
                    for ix, (_, child) in enumerate(self._families)
                )
        return istr + label_rep + families_rep

    def __repr__(self) -> str:
        """ Create a reasonably nice text representation of this node
            and its families of children, if any """
        return self._repr(0)

    def __str__(self) -> str:
        """ Return a string representation of this node """
        return "<Node: " + str(self._nonterminal or self._token) + ">"


class ParseError(Exception):

    """ Exception class for parser errors """

    def __init__(
        self, txt: str, token_index: Optional[int] = None, info: Any = None
    ) -> None:
        """ Store an information object with the exception,
            containing the parser state immediately before the error """
        super().__init__(txt)
        self._info = info
        self._token_index = token_index

    @property
    def info(self) -> Any:
        """ Return the parser state information object """
        return self._info

    @property
    def token_index(self) -> Optional[int]:
        """ Return the 0-based index of the token where the parser ran out of options """
        return self._token_index

    def __str__(self) -> str:
        """ Return a string representation of the parse error """
        return self.args[0]


class Fast_Parser(BIN_Parser):

    """ This class wraps an Earley-Scott parser written in C++,
        which is called via CFFI.

        The class supports the context manager protocol so you can say:

        with Fast_Parser() as fast_p:
           node = fast_p.go(...)

        C++ objects associated with the parser will then be cleaned
        up automatically upon exit of the context, whether by normal
        means or as a consequence of an exception.

        Otherwise, i.e. if not using a context manager, call fast_p.cleanup()
        after using the fast_p parser instance, preferably in a try/finally block.
    """

    # The C++ grammar object (a binary blob)
    _c_grammar: Any = ffi_NULL
    # The C++ grammar timestamp
    _c_grammar_ts: Optional[float] = None

    @classmethod
    def _load_binary_grammar(cls) -> Any:
        """ Load the binary grammar file into memory, if required """
        fname = cls._GRAMMAR_BINARY_FILE
        try:
            ts = os.path.getmtime(fname)
        except os.error:
            raise GrammarError("Binary grammar file {0} not found".format(fname))
        if cls._c_grammar == ffi_NULL or cls._c_grammar_ts != ts:
            # Need to load or reload the grammar
            if cls._c_grammar != ffi_NULL:
                # Delete previous grammar instance, if any
                eparser.deleteGrammar(cls._c_grammar)  # type: ignore
                cls._c_grammar = ffi_NULL
            cls._c_grammar = eparser.newGrammar(fname.encode("utf-8"))  # type: ignore
            cls._c_grammar_ts = ts
            if cls._c_grammar == ffi_NULL:
                raise GrammarError(
                    "Unable to load binary grammar file {0}".format(fname)
                )
        return cast(Any, cls._c_grammar)

    def __init__(self, verbose: bool = False, root: Optional[str] = None) -> None:

        # Only one initialization at a time, since we don't want a race
        # condition between threads with regards to reading and parsing the grammar file
        # vs. writing the binary grammar
        with GlobalLock("grammar"):
            # Read and parse the grammar text file
            super().__init__(verbose)
            # Create instances of the C++ Grammar and Parser classes
            c_grammar = self._load_binary_grammar()
            # Create a C++ parser object for the grammar, passing the proxies for the
            # two Python callback functions into it
            self._c_parser: Any = eparser.newParser(  # type: ignore
                c_grammar, eparser.matching_func, eparser.alloc_func  # type: ignore
            )
            # Find the index of the default root nonterminal for this parser instance
            self._root_index = (
                0 if root is None else self.grammar.nonterminals[root].index
            )
            # Maintain a token/terminal matching cache for the duration
            # of this parser instance. Note that this cache will grow with use,
            # as it includes an entry (consisting of one byte per terminal in the
            # grammar, or currently about 5K bytes for Greynir.grammar) for every
            # distinct token that the parser encounters.
            self._matching_cache: Dict[Tuple[Hashable, ...], Any] = dict()

    def __enter__(self):
        """ Python context manager protocol """
        return self

    # noinspection PyUnusedLocal
    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any):
        """ Python context manager protocol """
        self.cleanup()
        return False

    def go(self, tokens: Iterable[Tok], *, root: Optional[str] = None) -> Node:
        """ Call the C++ parser module to parse the tokens. The parser's
            default root nonterminal can be overridden by passing its
            name in the root parameter. """

        wrapped_tokens = self._wrap(tokens)  # Inherited from BIN_Parser
        lw = len(wrapped_tokens)
        err: Any = cast(Any, ffi).new("unsigned int*")
        result = None

        # Use the context manager protocol to guarantee that the parse job
        # handle will be properly deleted even if an exception is thrown

        with ParseJob.make(
            self.grammar, wrapped_tokens, self._terminals, self._matching_cache
        ) as job:

            # Determine the root nonterminal to be used for this parse
            if root is None:
                # Use the parser's default root
                root_index = self._root_index
            else:
                # Override the default root for this parse
                root_index = self.grammar.nonterminals[root].index

            node: Any = eparser.earleyParse(self._c_parser, lw, root_index, job.handle, err)  # type: ignore

            if node == ffi_NULL:
                ix = err[0]  # Token index
                if ix >= 1:
                    # Find the error token index in the original (unwrapped) token list
                    orig_ix = wrapped_tokens[ix].index if ix < lw else ix
                    raise ParseError(
                        "No parse available at token {0} ({1})".format(
                            orig_ix, wrapped_tokens[ix - 1]
                        ),
                        orig_ix - 1,
                    )
                else:
                    # Not a normal parse error, but report it anyway
                    raise ParseError(
                        "No parse available at token {0} ({1} tokens in input)".format(
                            ix, len(wrapped_tokens)
                        ),
                        0,
                    )

            # eparser.dumpForest(node, Fast_Parser._c_grammar) # !!! DEBUG
            # Create a new Python-side node forest corresponding to the C++ one
            result = Node.from_c_node(job, node)

        # Delete the C++ nodes
        eparser.deleteForest(node)  # type: ignore
        assert result is not None
        return result

    def go_no_exc(self, tokens: Iterable[Tok], **kwargs: Any) -> Optional[Node]:
        """ Simple version of go() that returns None instead of throwing ParseError """
        try:
            return self.go(tokens, **kwargs)
        except ParseError:
            return None

    def cleanup(self) -> None:
        """ Delete C++ objects. Must call after last use of Fast_Parser
            to avoid memory leaks. The context manager protocol is recommended
            to guarantee cleanup. """
        if self._c_parser != ffi_NULL:
            eparser.deleteParser(self._c_parser)  # type: ignore
        self._c_parser = ffi_NULL
        if Settings.DEBUG:
            eparser.printAllocationReport()  # type: ignore
            print(
                "Matching cache contains {0} entries".format(len(self._matching_cache))
            )

    @classmethod
    def discard_grammar(cls) -> None:
        """ Discard the C grammar object instance held as a class attribute """
        if cls._c_grammar != ffi_NULL:
            eparser.deleteGrammar(cls._c_grammar)  # type: ignore
        cls._c_grammar = ffi_NULL
        cls._c_grammar_ts = None

    @classmethod
    def num_combinations(cls, forest: Node) -> int:
        """ Count the number of possible parse tree combinations in the given forest """

        nc: Dict[Node, int] = dict()
        mul = operator.mul

        def _num_comb(w: Node) -> int:
            if w is None or w._token is not None:
                # Empty (epsilon) node or token node
                return 1
            # If a subtree has already been counted, re-use that count
            # (this is less efficient for small trees but much more efficient
            # for extremely ambiguous trees, with combinations in the
            # millions)
            cnt = nc.get(w)
            if cnt is not None:
                assert cnt is not NotImplemented, "Loop in node tree at {0}".format(w)
                return cnt
            nc[w] = NotImplemented  # Special marker for an unassigned cache entry
            comb = 0
            for _, f in w.enum_children():
                comb += reduce(mul, (_num_comb(ch) for ch in f if ch is not None), 1)
            result = nc[w] = comb if comb > 0 else 1
            return result

        return _num_comb(forest)


class ParseForestNavigator:

    """ Base class for navigating parse forests. Override the underscored
        methods to perform actions at the corresponding points of navigation. """

    # pylint: disable=assignment-from-none

    def __init__(self, visit_all: bool = False) -> None:
        """ If visit_all is False, we only visit each packed node once.
            If True, we visit the entire tree in order. """
        self._visit_all = visit_all

    def visit_epsilon(self, level: int) -> Any:
        """ At Epsilon node """
        return None

    def visit_token(self, level: int, w: Node) -> Any:
        """ At token node """
        return None

    def visit_nonterminal(self, level: int, node: Node) -> Any:
        """ At nonterminal node """
        # Typically returns an accumulation object to collect results
        return None

    def visit_family(
        self, results: Any, level: int, w: Node, ix: int, prod: Production
    ) -> None:
        """ At a family of children """
        return

    def add_result(self, results: Any, ix: int, r: Any) -> None:
        """ Append a single result r to the results accumulation object """
        return

    def process_results(self, results: Any, node: Node) -> Any:
        """ Process results after visiting children.
            The results list typically contains tuples (ix, r) where ix is
            the family index and r is the child result """
        return None

    def force_visit(
        self, w: Optional[Node], visited: Dict[Optional[Node], Any]
    ) -> bool:
        """ Override this and return True to visit a node, even if self._visit_all
            is False and the node has been visited before """
        return False

    def go(self, root_node: Node) -> Any:
        """ Navigate the forest from the root node """

        # Memoization cache dictionary
        visited: Dict[Optional[Node], Any] = dict()

        def _nav_helper(w: Optional[Node], level: int) -> Any:
            """ Navigate from w """
            if (
                not self._visit_all
                and w in visited
                and not self.force_visit(w, visited)
            ):
                # Already seen: return the memoized result
                return visited[w]
            if w is None:
                # Epsilon node
                v = self.visit_epsilon(level)
            elif w._token is not None:
                # Return the score of this terminal option
                v = self.visit_token(level, w)
            else:
                # Init accumulator for child results
                results = self.visit_nonterminal(level, w)
                if results is NotImplemented:
                    # If visit_nonterminal() returns NotImplemented,
                    # don't bother visiting children or processing
                    # results; instead _nav_helper() returns NotImplemented
                    v = results
                else:
                    if w._families:
                        if w.is_interior and not w.is_ambiguous:
                            child_level = level
                        else:
                            child_level = level + 1
                        for ix, (prod, children) in enumerate(w._families):
                            self.visit_family(results, level, w, ix, prod)
                            for ch in children:
                                self.add_result(
                                    results, ix, _nav_helper(ch, child_level)
                                )
                    v = self.process_results(results, w)
            if not self._visit_all:
                # Mark the node as visited and store its result
                visited[w] = v
            return v

        return _nav_helper(root_node, 0)


class ParseForestPrinter(ParseForestNavigator):

    """ Print a parse forest to stdout or a file """

    def __init__(
        self,
        detailed: bool = False,
        file: Optional[IO[str]] = None,
        show_scores: bool = False,
        show_ids: bool = False,
        visit_all: bool = True,
        skip_duplicates: bool = False,
    ) -> None:

        # Normally, we visit all nodes, also those we've seen before
        super().__init__(visit_all=visit_all)
        self._detailed = detailed
        self._file = file
        self._show_scores = show_scores
        self._show_ids = show_ids
        self._skip_duplicates = skip_duplicates
        self._visited: Set[Node] = set()

    def _score(self, w: Node) -> str:
        """ Return a string showing the node's score """
        # !!! To enable this, assignment of the .score attribute
        # !!! needs to be uncommented in reducer.py
        return " [{0}]".format(w.score) if self._show_scores else ""

    def visit_epsilon(self, level: int) -> None:
        """ Epsilon (null) node """
        indent = "  " * level  # Two spaces per indent level
        print(indent + "(empty)", file=self._file)

    def visit_token(self, level: int, w: Node) -> None:
        """ Token matching a terminal """
        indent = "  " * level  # Two spaces per indent level
        h = str(w.token)
        if self._show_ids:
            h += " @ {0:x}".format(id(w))
        print(
            indent + "{0}: {1}{2}".format(w.terminal, h, self._score(w)),
            file=self._file,
        )

    def visit_nonterminal(self, level: int, node: Node) -> Any:
        # Interior nodes are not printed
        # and do not increment the indentation level
        if self._detailed or not node.is_interior:
            nt = cast(Nonterminal, node.nonterminal)
            if not self._detailed:
                if node.is_empty and nt.is_optional:
                    # Skip printing optional nodes that don't contain anything
                    return NotImplemented  # Don't visit child nodes
            h = nt.name
            indent = "  " * level  # Two spaces per indent level
            if self._show_ids:
                h += " @ {0:x}".format(id(node))
            print(indent + h + self._score(node), file=self._file)
            if self._skip_duplicates:
                # We don't want to redisplay entire subtrees that we've
                # seen before
                if node in self._visited:
                    print(indent + "  <Seen before>", file=self._file)
                    return NotImplemented  # Don't visit child nodes
                self._visited.add(node)
        return None  # No results required, but visit children

    def visit_family(
        self, results: Any, level: int, w: Node, ix: int, prod: Production
    ) -> None:
        """ Show trees for different options, if ambiguous """
        if w.is_ambiguous:
            indent = "  " * level  # Two spaces per indent level
            print(indent + "Option " + str(ix + 1) + ":", file=self._file)

    @classmethod
    def print_forest(
        cls,
        root_node: "Node",
        detailed: bool=False,
        file: Optional[IO[str]]=None,
        show_scores: bool=False,
        show_ids: bool=False,
        visit_all: bool=True,
        skip_duplicates: bool=False,
    ):
        """ Print a parse forest to the given file, or stdout if none """
        cls(
            detailed,
            file,
            show_scores,
            show_ids,
            visit_all,
            skip_duplicates=skip_duplicates,
        ).go(root_node)


class ParseForestDumper(ParseForestNavigator):

    """ Dump a parse forest into a compact string """

    # The result is a string consisting of lines separated by newline characters.
    # The format is as follows:
    # (n indicates a nesting level, >= 0)
    # R1 -- start indicator and version number
    # Cn -- the sentence's parse tree score
    # Ln -- the sentence length in tokens
    # Pn -- Epsilon node
    # Tn terminal token -- Token/terminal node
    # Nn nonterminal -- Nonterminal node
    # On index -- Option with index >= 0
    # Q0 -- end indicator (not followed by newline)

    def __init__(self, token_dicts: Optional[List[TokenDict]]) -> None:
        super().__init__(visit_all=True)  # Visit all nodes
        self._result = ["R1"]  # Start indicator and version number
        self._token_dicts = token_dicts

    def visit_epsilon(self, level: int) -> Any:
        # Identify this as an epsilon (null) node
        # !!! Not necessary - removed July 2018 VTh
        # self._result.append("P{0}".format(level))
        return None

    def visit_token(self, level: int, w: Node) -> Any:
        # Identify this as a terminal/token
        ta = ""  # Augmented terminal
        assert w.token is not None
        assert w.terminal is not None
        name = w.terminal.name
        if self._token_dicts is not None:
            # Get the descriptor dict for this token/terminal match
            td = self._token_dicts[w.token.index]
            if "t" in td and "m" in td:
                # Calculate an augmented terminal, including additional info from BÍN
                if td["t"] != name:
                    assert False
                ta = simplify_terminal(td["t"], td["m"][1])  # Fallback category
                # The m(3) field is 'beyging'
                ta = augment_terminal(ta, td["x"].lower(), td["m"][3])
                if name == ta:
                    ta = ""  # No need to repeat augmented terminal if it is identical
                else:
                    ta = " " + ta

        self._result.append("T{0} {1} {2}{3}".format(level, name, w.token.dump, ta))
        return None

    def visit_nonterminal(self, level: int, node: Node) -> Any:
        # Interior nodes are not dumped
        # and do not increment the indentation level
        if not node.is_interior:
            nt = node.nonterminal
            assert nt is not None
            if node.is_empty and nt.is_optional:
                # Skip printing optional nodes that don't contain anything
                return NotImplemented  # Don't visit child nodes
            # Identify this as a nonterminal
            self._result.append("N{0} {1}".format(level, nt.name))
        return None  # No results required, but visit children

    def visit_family(
        self, results: Any, level: int, w: Node, ix: int, prod: Production
    ) -> None:
        if w.is_ambiguous:
            # Identify this as an option
            self._result.append("O{0} {1}".format(level, ix))

    @classmethod
    def dump_forest(
        cls, root_node: Node, token_dicts: Optional[List[TokenDict]] = None
    ) -> str:
        """ Return a string with a multi-line text representation of the parse tree """
        dumper = cls(token_dicts)
        dumper.go(root_node)
        dumper._result.append("Q0")  # End marker
        return "\n".join(dumper._result)


class _FlattenerNode:

    """ A node in a flattened parse tree, produced by
        the ParseTreeFlattener class (below) """

    def __init__(self, p: FlattenerType, score: int) -> None:
        self._p = p
        self._score = score
        self._children: Optional[List[_FlattenerNode]] = None

    def add_child(self, child: "_FlattenerNode") -> None:
        if self._children is None:
            self._children = [child]
        else:
            self._children.append(child)

    @property
    def p(self) -> FlattenerType:
        return self._p

    @property
    def children(self) -> Optional[List["_FlattenerNode"]]:
        return self._children

    @property
    def has_children(self) -> bool:
        return self._children is not None

    @property
    def is_nonterminal(self) -> bool:
        return not isinstance(self._p, tuple)

    @property
    def score(self) -> int:
        return self._score

    def _to_str(self, indent: int) -> str:
        if self._children is not None:
            return "{0}{1}{2}".format(
                " " * indent,
                self._p,
                "".join("\n" + child._to_str(indent + 1) for child in self._children),
            )
        return "{0}{1}".format(" " * indent, self._p)

    def __str__(self) -> str:
        return self._to_str(0)


class ParseForestFlattener(ParseForestNavigator):

    """ Create a simpler, flatter version of an already disambiguated parse tree """

    def __init__(self) -> None:
        super().__init__(visit_all=True)  # Visit all nodes
        self._stack: Optional[List[_FlattenerNode]] = None

    def go(self, root_node: Node) -> None:
        self._stack = None
        super().go(root_node)

    @property
    def root(self) -> Optional[_FlattenerNode]:
        return self._stack[0] if self._stack else None

    def visit_epsilon(self, level: int) -> Any:
        """ Epsilon (null) node: not included in a flattened tree """
        return None

    def visit_token(self, level: int, w: Node) -> Any:
        """ Add a terminal/token node to the flattened tree """
        # assert level > 0
        # assert self._stack
        assert w.terminal is not None
        assert w.token is not None
        node = _FlattenerNode((w.terminal, w.token), w.score)
        assert self._stack is not None
        self._stack = self._stack[0:level]
        self._stack[-1].add_child(node)
        return None

    def visit_nonterminal(self, level: int, node: Node) -> Any:
        """ Add a nonterminal node to the flattened tree """
        # Interior nodes are not dumped
        # and do not increment the indentation level
        if not node.is_interior:
            assert node.nonterminal is not None
            if node.is_empty and node.nonterminal.is_optional:
                # Skip optional nodes that don't contain anything
                return NotImplemented  # Signal: Don't visit child nodes
            # Identify this as a nonterminal
            fnode = _FlattenerNode(node.nonterminal, node.score)
            if level == 0:
                # New root (must be the only one)
                assert self._stack is None
                self._stack = [fnode]
            else:
                # New child of the parent node
                assert self._stack is not None
                self._stack = self._stack[0:level]
                self._stack[-1].add_child(fnode)
                self._stack.append(fnode)
        return None  # No results required, but visit children

    def visit_family(
        self, results: Any, level: int, w: Node, ix: int, prod: Production
    ) -> None:
        """ Visit different subtree options within a parse forest """
        # In this case, the tree should be unambigous
        assert not w.is_ambiguous

    @classmethod
    def flatten(cls, root_node: Node) -> Optional[_FlattenerNode]:
        """ Flatten a parse tree """
        dumper = cls()
        dumper.go(root_node)
        return dumper.root



"""

    Greynir: Natural language processing for Icelandic

    GlobalLock utility class

    Copyright (C) 2021 Miðeind ehf.
    Original author: Vilhjálmur Þorsteinsson

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    This module implements the GlobalLock class, providing
    interprocess locks within a server.

    A GlobalLock is implemented as a file in the /tmp/ directory,
    which is assumed to exist (on the current drive in Windows).

    A quick and easy way to use a blocking GlobalLock is as follows:

    with GlobalLock('somestring'):
        code_that_only_one_process_can_run_simultaneously()

"""

from typing import Any, IO, Optional

import os
import stat
import tempfile


class LockError(Exception):
    """ Lock could not be obtained """
    pass


POSIX = False

try:
    # Try Linux/POSIX
    import fcntl
except ImportError:

    try:
        # Try Windows
        import msvcrt
    except ImportError:

        # Not Unix, not Windows: bail out
        def _lock_file(file: IO[str], block: bool) -> None:
            raise TypeError("File locking not supported on this platform")

        def _unlock_file(file: IO[str]) -> None:
            raise TypeError("File locking not supported on this platform")

    else:

        # Windows

        def _lock_file(file: IO[str], block: bool) -> None:  # type: ignore
            # Lock just the first byte of the file
            retry = True
            while retry:
                retry = False
                try:
                    msvcrt.locking(  # type: ignore
                        file.fileno(),
                        msvcrt.LK_LOCK if block else msvcrt.LK_NBLCK,  # type: ignore
                        1
                    )
                except OSError as e:
                    if block and e.errno == 36:
                        # Windows says 'resource deadlock avoided', but we truly want
                        # a longer blocking wait: try again
                        retry = True
                    else:
                        raise LockError(
                            "Couldn't lock {0}, errno is {1}".format(file.name, e.errno)
                        )

        def _unlock_file(file: IO[str]) -> None:  # type: ignore
            try:
                file.seek(0)
                msvcrt.locking(file.fileno(), msvcrt.LK_UNLCK, 1)  # type: ignore
            except OSError as e:
                raise LockError(
                    "Couldn't unlock {0}, errno is {1}".format(file.name, e.errno)
                )

else:

    # Linux/POSIX

    POSIX = True

    def _lock_file(file: IO[str], block: bool) -> None:  # type: ignore
        try:
            fcntl.flock(file.fileno(), fcntl.LOCK_EX | (0 if block else fcntl.LOCK_NB))
        except IOError:
            raise LockError("Couldn't lock {0}".format(file.name))

    def _unlock_file(file: IO[str]) -> None:  # type: ignore
        # File is automatically unlocked on close
        pass


class GlobalLock:

    _TMP_DIR = tempfile.gettempdir()

    def __init__(self, lockname: str) -> None:
        """ Initialize a global lock with the given name """
        assert lockname and isinstance(lockname, str)
        # Locate global locks in the system temporary directory
        # (should work on both Windows and Unix/POSIX)
        self._path = os.path.join(self._TMP_DIR, "greynir-" + lockname)
        self._fp: Optional[IO[str]] = None

    def acquire(self, block: bool=True) -> None:
        """ Acquire a global lock, blocking if block = True """

        if self._fp is not None:
            # Already hold the lock
            return

        path = self._path
        fp = None
        try:
            # Try to open for writing without truncation:
            fp = open(path, "r+")
        except IOError:
            # If the file doesn't exist, we'll get an IO error, try a+
            # Note that there may be a race here. Multiple processes
            # could fail on the r+ open and open the file a+, but only
            # one will get the the lock and write a pid.
            fp = open(path, "a+")
            # Make sure that the file is readable and writable by others
            if POSIX and fp is not None:
                os.fchmod(
                    fp.fileno(),
                    stat.S_IRUSR | stat.S_IWUSR
                    | stat.S_IRGRP | stat.S_IWGRP
                    | stat.S_IROTH | stat.S_IWOTH
                )

        if fp is None:
            raise LockError("Couldn't open or create lock file {0}".format(path))

        self._fp = fp

        try:
            _lock_file(fp, block)
        except:
            fp.seek(1)
            fp.close()
            raise

        # Once acquired, write the process id to the file
        fp.write(" %s\n" % os.getpid())
        fp.truncate()
        fp.flush()

    def release(self) -> None:
        """ Release the lock """
        if self._fp is not None:
            _unlock_file(self._fp)
            self._fp.close()
            self._fp = None

    def __enter__(self):
        """ Python context manager protocol """
        self.acquire(block=True)
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any):
        """ Python context manager protocol """
        self.release()
        return False



"""

	cache.py

    Cache utility classes

    The LRU_Cache and LFU_Cache classes herein are
	copyright (C) 2011 by Raymond Hettinger

    cf. http://code.activestate.com/recipes/577970-simplified-lru-cache/
        http://code.activestate.com/recipes/498245-lru-and-lfu-cache-decorators/

    MIT license:

    Permission is hereby granted, free of charge, to any person obtaining a copy of
    this software and associated documentation files (the "Software"), to deal in
    the Software without restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
    Software, and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.

    ---

    The classes have been modified from their original versions,
    which are available from the URLs given above.

"""

from typing import List, Dict, Any, Callable, TypeVar, Generic, cast

from heapq import nsmallest
from operator import itemgetter
import threading
from functools import wraps


LRU_DEFAULT = 1024
LFU_DEFAULT = 512


_K = TypeVar("_K")
_V = TypeVar("_V")


class LRU_Cache(Generic[_V]):

    def __init__(
        self, user_function: Callable[..., _V], maxsize: int=LRU_DEFAULT
    ) -> None:
        # Link layout:     [PREV, NEXT, KEY, RESULT]
        root: List[Any] = [None, None, None, None]
        self.root = root
        self.user_function = user_function
        cache: Dict[Any, List[Any]] = {}
        self.cache = cache

        last = root
        for _ in range(maxsize):
            key = object()
            cache[key] = last[1] = last = [last, root, key, None]
        root[0] = last

    def __call__(self, *key: Any) -> _V:
        cache = self.cache
        root = self.root
        link = cache.get(key)
        if link is not None:
            link_prev, link_next, _, result = link
            link_prev[1] = link_next
            link_next[0] = link_prev
            last = root[0]
            last[1] = root[0] = link
            link[0] = last
            link[1] = root
            return result
        result = self.user_function(*key)
        root[2] = key
        root[3] = result
        oldroot = root
        root = self.root = root[1]
        root[2], oldkey = None, root[2]
        root[3] = None
        del cache[oldkey]
        cache[key] = oldroot
        return result


class Counter(Dict[_K, int], Generic[_K]):
    """ Mapping where default values are zero """
    def __missing__(self, key: _K) -> int:
        return 0


class LFU_Cache(Generic[_K, _V]):

    """ Least-frequently-used (LFU) cache for word lookups.
        Based on a pattern by Raymond Hettinger
    """

    def __init__(self, maxsize: int=LFU_DEFAULT) -> None:
        # Mapping of keys to results
        self.cache: Dict[_K, _V] = {}
        # Times each key has been accessed
        self.use_count: Counter[_K] = Counter()
        self.maxsize = maxsize
        self.hits = self.misses = 0
        # The cache may be accessed in parallel by multiple threads
        self.lock = threading.Lock()

    def lookup(self, key: _K, func: Callable[[_K], _V]) -> _V:
        """ Lookup a key in the cache, calling func(key)
            to obtain the data if not already there """
        with self.lock:
            self.use_count[key] += 1
            # Get cache entry or compute if not found
            try:
                result = self.cache[key]
                self.hits += 1
            except KeyError:
                result = func(key)
                self.cache[key] = result
                self.misses += 1

                # Purge the 10% least frequently used cache entries
                if len(self.cache) > self.maxsize:
                    for key, _ in nsmallest(self.maxsize // 10,
                        self.use_count.items(), key = itemgetter(1)):

                        del self.cache[key], self.use_count[key]

            return result


# Define a type variable to allow MyPy to infer the relationship
# between intermediate types in cached and cached_property
_T = TypeVar('_T')

# Define a unique singleton for use as a sentinel
_NA = object()


def cached(func: Callable[..., _T]) -> Callable[..., _T]:
    """ A decorator for caching function calls """
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> _T:
        val = cast(_T, getattr(func, "_cache", _NA))
        if val is _NA:
            val = func(*args, **kwargs)
            setattr(func, "_cache", val)
        return val
    return wrapper


class cached_property(Generic[_T]):

    """ A decorator for caching instance properties """

    def __init__(self, func: Callable[..., _T]) -> None:
        self.__doc__ = getattr(func, "__doc__")
        self.func = func

    def __get__(self, obj: Any, cls: Any) -> _T:
        if obj is None:
            return cast(_T, self)  # Hack to satisfy mypy/Pylance
        # Get the property value and put it into the instance's
        # dict instead of the original function
        val = obj.__dict__[self.func.__name__] = self.func(obj)
        return val



"""

    Greynir: Natural language processing for Icelandic

    Copyright (C) 2021 Miðeind ehf.
    Original author: Vilhjálmur Þorsteinsson

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    This module exposes the Greynir API, i.e. the identifiers that are
    directly accessible via the reynir module object after importing it.

"""

# Expose the Greynir API

from .reynir import (
    Greynir,
    Reynir,
    Terminal,
    LemmaTuple,
    ProgressFunc,
    ParseResult,
    Sentence,
    Paragraph,
    ICELANDIC_RATIO,
)

# Import the following _underscored classes to be able to use them
# in type signatures in derived classes
from .reynir import (
    _Job,
    _Sentence,
    _Paragraph,
)
from .nounphrase import NounPhrase
from .fastparser import ParseForestPrinter, ParseForestDumper, ParseForestFlattener
from .fastparser import ParseError, ParseForestNavigator
from .settings import Settings
from .bintokenizer import tokenize, TokenList
from .version import __version__

# Expose the tokenizer API

from tokenizer import (
    TOK,
    Tok,
    paragraphs,
    correct_spaces,
    mark_paragraphs,
)
from tokenizer.abbrev import Abbreviations
from tokenizer import (
    TP_LEFT,
    TP_CENTER,
    TP_RIGHT,
    TP_NONE,
    TP_WORD,
    KLUDGY_ORDINALS_PASS_THROUGH,
    KLUDGY_ORDINALS_MODIFY,
    KLUDGY_ORDINALS_TRANSLATE,
)

__all__ = (
    "TP_LEFT", "TP_RIGHT", "TP_CENTER", "TP_NONE", "TP_WORD",
    "KLUDGY_ORDINALS_MODIFY", "KLUDGY_ORDINALS_PASS_THROUGH",
    "KLUDGY_ORDINALS_TRANSLATE",
    "Greynir",
    "Reynir",
    "Terminal",
    "LemmaTuple",
    "ProgressFunc",
    "ParseResult",
    "Sentence",
    "Paragraph",
    "ICELANDIC_RATIO",
    "TOK", "Tok",
    "paragraphs",
    "correct_spaces",
    "mark_paragraphs",
    "_Job",
    "_Sentence",
    "_Paragraph",
    "NounPhrase",
    "ParseForestPrinter", "ParseForestDumper", "ParseForestFlattener",
    "ParseError", "ParseForestNavigator",
    "Settings",
    "tokenize", "TokenList",
)

__author__ = "Miðeind ehf."
__copyright__ = "(C) 2021 Miðeind ehf."

Abbreviations.initialize()
Settings.read("config/GreynirPackage.conf")




"""

    Greynir: Natural language processing for Icelandic

    Copyright (C) 2021 Miðeind ehf.

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    This module contains a function to (simplistically) lemmatize text
    without parsing it.

"""

from typing import Optional, Union, Callable, Tuple, List, Iterator, TypeVar, cast

from abc import abstractmethod, ABCMeta
from collections import OrderedDict

from .bindb import BIN_Tuple
from .bintokenizer import tokenize, TOK


# In Python >= 3.8, the base class could be typing.Protocol
class Comparable(metaclass=ABCMeta):

    """ Protocol for annotating comparable types """

    @abstractmethod
    def __lt__(self: "CT", other: "CT") -> bool:
        ...


CT = TypeVar("CT", bound=Comparable)

LemmaTuple = Tuple[str, str]  # Lemma, category (ordfl)


def simple_lemmatize(
    txt: str,
    *,
    all_lemmas: bool = False,
    sortkey: Optional[Callable[[LemmaTuple], Comparable]] = None,
) -> Union[Iterator[LemmaTuple], Iterator[List[LemmaTuple]]]:
    """ Simplistically lemmatize a list of tokens, returning a generator of
        (lemma, category) tuples. The default behaviour is to return the
        first lemma provided by bintokenizer. If all_lemmas are requested,
        returns full list of potential lemmas. A sort function can be provided
        to determine the ordering of that list. """
    for t in tokenize(txt):
        y: Optional[List[LemmaTuple]] = None
        if t.kind == TOK.WORD:
            if t.val:
                # Known word
                if "-" in t.txt:
                    # The original word already contains a hyphen: leave'em in
                    y = [(v.stofn, v.ordfl) for v in cast(List[BIN_Tuple], t.val)]
                else:
                    # The original word doesn't contain a hyphen: any hyphens
                    # in the lemmas must come from the compounding algorithm
                    y = [(v.stofn.replace("-", ""), v.ordfl) for v in cast(List[BIN_Tuple], t.val)]
            else:
                # Unknown word: assume it's an entity
                y = [(t.txt, "entity")]
        elif t.kind == TOK.PERSON:
            assert t.person_names
            # Person name w. gender
            person_name = t.person_names[0]
            y = [(person_name.name, "person_" + (person_name.gender or "hk"))]
        elif t.kind == TOK.ENTITY or t.kind == TOK.COMPANY:
            # Entity or company name
            y = [(t.txt, "entity")]
        if y is not None:
            # OK, we're returning one or more lemmas for this token
            # Remove duplicates while preserving order
            y = list(OrderedDict.fromkeys(y))
            if sortkey is not None:
                y.sort(key=sortkey)
            if all_lemmas:
                yield y
            else:
                yield y[0]  # Naively return first lemma



"""

    Greynir: Natural language processing for Icelandic

    BinDb module

    Copyright (C) 2021 Miðeind ehf.

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    This module implements a thin wrapper on top of the GreynirBin
    class from BinPackage, as well as a couple of basic data classes.

"""

from typing import Any, List, Optional, Tuple
from functools import lru_cache

from islenska.basics import make_bin_entry
from islenska.bindb import GreynirBin as GBin, PERSON_NAME_FL

from tokenizer import BIN_Tuple

from .settings import StaticPhrases

# SHSnid tuple as seen by the Greynir compatibility layer
ResultTuple = Tuple[str, List[BIN_Tuple]]


# Size of name cache for lookup_name_gender
_NAME_GENDER_CACHE_SIZE = 128


class GreynirBin(GBin):

    """ Overridden class that adds a singleton instance of GreynirBin
        and a context manager protocol """

    _singleton: Optional["GreynirBin"] = None

    @classmethod
    def get_db(cls) -> "GreynirBin":
        if cls._singleton is None:
            cls._singleton = GreynirBin()
        return cls._singleton

    def __enter__(self) -> "GreynirBin":
        """ Allow this class to be used in a with statement """
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        pass

    def lookup_g(
        self, w: str, at_sentence_start: bool = False, auto_uppercase: bool = False
    ) -> ResultTuple:
        """ Returns BIN_Tuple instances, which are the Greynir version
            of islenska.BinEntry """
        w, m = self._lookup(
            w,
            at_sentence_start,
            auto_uppercase,
            self._meanings_cache_lookup,
            make_bin_entry,
        )
        return w, [BIN_Tuple._make(mm) for mm in m]

    def lookup_nominative_g(self, w: str, **options: Any) -> List[BIN_Tuple]:
        """ Returns the Greynir version of islenska.BinEntry """
        return [BIN_Tuple._make(mm) for mm in super().lookup_nominative(w, **options)]

    def lookup_accusative_g(self, w: str, **options: Any) -> List[BIN_Tuple]:
        """ Returns the Greynir version of islenska.BinEntry """
        return [BIN_Tuple._make(mm) for mm in super().lookup_accusative(w, **options)]

    def lookup_dative_g(self, w: str, **options: Any) -> List[BIN_Tuple]:
        """ Returns the Greynir version of islenska.BinEntry """
        return [BIN_Tuple._make(mm) for mm in super().lookup_dative(w, **options)]

    def lookup_genitive_g(self, w: str, **options: Any) -> List[BIN_Tuple]:
        """ Returns the Greynir version of islenska.BinEntry """
        return [BIN_Tuple._make(mm) for mm in super().lookup_genitive(w, **options)]

    def meanings(self, w: str) -> List[BIN_Tuple]:
        """ Low-level lookup of BIN_Tuple instances for the given word """
        return [
            BIN_Tuple(k.ord, k.bin_id, k.ofl, k.hluti, k.bmynd, k.mark)
            for k in self._ksnid_lookup(w)
        ]

    @lru_cache(maxsize=_NAME_GENDER_CACHE_SIZE)
    def lookup_name_gender(self, name: str) -> str:
        """ Given a person name, lookup its gender """
        if not name:
            return "hk"  # Unknown gender
        w = name.split(maxsplit=1)[0]  # First name
        g = self.meanings(w)
        m = next((x for x in g if x.fl in PERSON_NAME_FL), None)
        if m:
            # Found a name meaning
            return m.ordfl
        # The first name was not found: check whether the full name is
        # in the static phrases
        m = StaticPhrases.lookup(name)
        if m is not None:
            if m.fl in PERSON_NAME_FL:
                return m.ordfl
        return "hk"  # Unknown gender



# type: ignore
"""

    test_serializers.py

    Tests for JSON serialization of sentences

    Copyright (C) 2021 by Miðeind ehf.

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

import json

import pytest  # type: ignore


@pytest.fixture(scope="module")
def r():
    """ Provide a module-scoped Greynir instance as a test fixture """
    from reynir import Greynir
    r = Greynir()
    yield r
    # Do teardown here
    r.__class__.cleanup()


def test_serializers(r):
    sents = [
        "Ég fór niðrá bryggjuna með Reyni Vilhjálmssyni í gær.",
        "Það var 17. júní árið 2020.",
        "Við sáum tvo seli og örugglega fleiri en 100 máva.",
        "Klukkan var orðin tólf þegar við fórum heim.",
        "Bíllinn kostaði €30.000 en ég greiddi 25500 USD fyrir hann.",
        "Morguninn eftir vaknaði ég kl. 07:30.",
        "Ég var fyrstur á fætur en Þuríður Hálfdánardóttir var númer 2.",
    ]
    for sent in sents:
        orig = r.parse_single(sent)
        assert orig.tree is not None

        json_str = r.dumps_single(orig, indent=2)
        new = r.loads_single(json_str)

        assert new.tree is not None

        assert all(ot.equal(nt) for ot, nt in zip(orig.tokens, new.tokens))
        assert orig.terminals == new.terminals

        assert orig.tree.flat_with_all_variants == orig.tree.flat_with_all_variants
        cls = r.__class__
        assert json.loads(orig.dumps(cls, indent=2)) == json.loads(new.dumps(cls, indent=2))


def test_annotree():
    s = """
            (META (ID-CORPUS 43bf66f3-51c4-11e6-8438-04014c605401.10)
            (ID-LOCAL greynir_corpus_00003.psd,.1)
            (URL http://www.mbl.is/sport/efstadeild/2016/07/24/ia_ibv_stadan_er_1_0/))
        (S0 (S-HEADING (IP (NP-SUBJ (fn_ft_kk_nf Engir (lemma enginn))
                                    (no_ft_kk_nf atburðir (lemma atburður)))
                            (NP-PRD (VP (so_ft_kk_lhþt_nf_sb skráðir (lemma skrá))))
                            (ADVP (ao enn (lemma enn))))))

    """
    from reynir.simpletree import AnnoTree
    atree = AnnoTree(s)
    stree = atree.as_simple_tree()
    assert stree is not None
    assert stree.text == "Engir atburðir skráðir enn"
    assert stree.tidy_text == "Engir atburðir skráðir enn"
    assert stree.nouns == ["atburður"]
    assert stree.verbs == ["skrá"]


if __name__ == "__main__":
    # When invoked as a main module, do a verbose test
    from reynir import Greynir
    g = Greynir()
    test_serializers(g)
    g.__class__.cleanup()



"""

    Greynir: Natural language processing for Icelandic

    Utility class for incremental parsing of token streams

    Copyright (C) 2021 Miðeind ehf.
    Original author: Vilhjálmur Þorsteinsson

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    This module implements a simple utility class for parsing token
    streams into paragraphs and sentences. The parse is incremental so
    that the client can take action on each paragraph and sentence as
    it is processed. Also, time.sleep(0) is called between sentences
    to make multi-threaded parses proceed more smoothly and evenly.

"""

import time
from typing import Iterable, Iterator, List, Optional, Tuple

from tokenizer import paragraphs, Tok

from .bintokenizer import tokens_are_foreign
from .fastparser import Fast_Parser, Node, ParseError
from .reducer import Reducer
from .settings import Settings


# Number of tree combinations that must be exceeded for a verbose
# parse dump to include the sentence text (as opposed to just basic stats)
_VERBOSE_AMBIGUITY_THRESHOLD = 1000

# The ratio of words in a sentence that must be found in BÍN
# for it to be analyzed as an Icelandic sentence
ICELANDIC_RATIO = 0.5


# The same type is defined in the Tokenizer module
SentenceTuple = Tuple[int, List[Tok]]


class IncrementalParser:

    """ Utility class to parse a token list as a sequence of paragraphs
        containing sentences. Typical usage:

        toklist = tokenize(text)
        fp = Fast_Parser()
        ip = IncrementalParser(fp, toklist)
        for p in ip.paragraphs():
            for sent in p.sentences():
                if sent.parse():
                    # sentence parsed successfully
                    # do something with sent.tree
                else:
                    # an error occurred in the parse
                    # the error token index is at sent.err_index
        num_sentences = ip.num_sentences
        num_parsed = ip.num_parsed
        ambiguity = ip.ambiguity
        parse_time = ip.parse_time

    """

    class _IncrementalSentence:

        """ An internal sentence representation class """

        def __init__(self, ip: "IncrementalParser", s: List[Tok]) -> None:
            self._ip = ip
            self._s = s
            self._len = len(s)
            assert self._len > 0  # Input should be already sanitized
            self._err_index: Optional[int] = None
            self._tree: Optional[Node] = None
            self._score = 0
            self._error: Optional[ParseError] = None

        def __len__(self):
            return self._len

        def parse(self) -> bool:
            """ Parse the sentence """
            num = 0
            score = 0
            forest: Optional[Node] = None
            try:
                if tokens_are_foreign(self._s, min_icelandic_ratio=ICELANDIC_RATIO):
                    raise ParseError(
                        "Sentence is probably not in Icelandic", token_index=0
                    )
                forest = self._ip._parser.go(self._s)
                if forest is not None:
                    num = Fast_Parser.num_combinations(forest)
                    if num > 1:
                        forest, score = self._ip._reducer.go_with_score(forest)
            except ParseError as e:
                # The ParseError may originate in the reducer.go_with_score()
                # function, and in that case, forest is not None; be sure to reset it
                forest = None
                score = 0
                num = 0
                self._err_index = e.token_index
                self._error = e
            self._tree = forest
            self._score = score
            self._ip._add_sentence(self, num)
            return num > 0

        @property
        def tokens(self) -> List[Tok]:
            return self._s

        @property
        def tree(self) -> Optional[Node]:
            return self._tree

        @property
        def score(self) -> int:
            return self._score

        @property
        def error(self) -> Optional[ParseError]:
            return self._error

        @property
        def err_index(self) -> int:
            return self._len - 1 if self._err_index is None else self._err_index

        @property
        def text(self) -> str:
            return " ".join(t.txt for t in self._s if t.txt)

        def __str__(self) -> str:
            return self.text

    class _IncrementalParagraph:

        """ An internal paragraph representation class """

        def __init__(self, ip: "IncrementalParser", p: List[SentenceTuple]) -> None:
            self._ip = ip
            self._p = p

        def sentences(self) -> Iterator["IncrementalParser._IncrementalSentence"]:
            """ Yield the sentences within the paragraph, nicely wrapped """
            Sent = IncrementalParser._IncrementalSentence
            for _, sent in self._p:
                # Call time.sleep(0) to yield the current thread, i.e.
                # enable the threading subsystem and/or eventlet under Gunicorn
                # to switch threads at this point - since the parsing of an
                # entire article can take a long time
                time.sleep(0)
                yield Sent(self._ip, sent)

    def __init__(self, parser: Fast_Parser, toklist: Iterable[Tok], verbose: bool=False) -> None:
        self._parser = parser
        self._reducer = Reducer(parser.grammar)
        self._num_sent = 0
        self._num_parsed_sent = 0
        self._num_tokens = 0
        self._num_combinations = 0
        self._total_score = 0
        self._total_ambig = 0.0
        self._total_tokens = 0
        self._start_time = self._last_time = time.time()
        self._verbose = verbose
        self._toklist = list(toklist)

    def _add_sentence(self, s: "IncrementalParser._IncrementalSentence", num: int) -> None:
        """ Add a processed sentence to the statistics """
        slen = len(s)
        self._num_sent += 1
        self._num_tokens += slen
        if num > 0:
            # The sentence was parsed successfully
            self._num_parsed_sent += 1
            self._num_combinations += num
            ambig_factor = num ** (1 / slen)
            self._total_ambig += ambig_factor * slen
            self._total_tokens += slen
            self._total_score += s.score
        # Debugging output, if requested and enabled
        if self._verbose and Settings.DEBUG:
            current_time = time.time()
            print(
                "Parsed sentence of length {0} with {1} combinations{3} "
                "in {4:.1f} seconds{2}".format(
                    slen,
                    num,
                    ("\n" + s.text) if num >= _VERBOSE_AMBIGUITY_THRESHOLD else "",
                    " and score " + str(s.score) if num >= 1 else "",
                    current_time - self._last_time,
                )
            )
            self._last_time = current_time

    def paragraphs(self) -> Iterator["IncrementalParser._IncrementalParagraph"]:
        """ Yield the paragraphs from the token stream """
        Para = IncrementalParser._IncrementalParagraph
        for p in paragraphs(self._toklist):
            yield Para(self, p)

    @property
    def num_tokens(self) -> int:
        return self._num_tokens

    @property
    def num_sentences(self) -> int:
        return self._num_sent

    @property
    def num_parsed(self) -> int:
        return self._num_parsed_sent

    @property
    def num_combinations(self) -> int:
        return self._num_combinations

    @property
    def total_score(self) -> int:
        return self._total_score

    @property
    def ambiguity(self) -> float:
        return (
            (self._total_ambig / self._total_tokens) if self._total_tokens > 0 else 1.0
        )

    @property
    def parse_time(self) -> float:
        return time.time() - self._start_time

